import ast
import inspect
import textwrap


class LogTransformer(ast.NodeTransformer):
    def visit_FunctionDef(self, node):
        # Create a logging statement
        log_statement = ast.Expr(
            value=ast.Call(
                func=ast.Name(id="print", ctx=ast.Load()),
                args=[ast.Constant(value=f"Entering function {node.name}")],
                keywords=[],
            )
        )

        # Set lineno and col_offset for the new statement
        log_statement.lineno = node.body[0].lineno
        log_statement.col_offset = node.body[0].col_offset

        # Insert the logging statement at the beginning of the function body
        node.body.insert(0, log_statement)
        return node


def wrap_with_logging(func):
    # Get the source code of the function
    source = inspect.getsource(func)
    source = textwrap.dedent(source)

    # Parse the source code into an AST
    tree = ast.parse(source)

    # Transform the AST to add logging
    transformer = LogTransformer()
    tree = transformer.visit(tree)

    # Fix the line numbers and column offsets
    ast.fix_missing_locations(tree)

    # Compile the modified AST
    code = compile(tree, filename="<ast>", mode="exec")

    # Extract the modified function from the compiled code
    namespace = {}
    exec(code, func.__globals__, namespace)

    # Return the modified function
    return namespace[func.__name__]


def assert_false():
    assert False


def wrap_id(func):
    return func


def wrap(func):
    func = wrap_id(func)
    func = wrap_with_logging(func)
    return func


# Define the function you want to wrap
@wrap
def my_function(x, y):
    false = assert_false()
    return x + y


# Wrap the function with logging
# my_function = wrap_with_logging(my_function)

# Call the modified function
result = my_function(2, 3)
print(result)



import functools

from oryx.core.interpreters.harvest import sow, harvest, reap, nest

import jax
import jax.lax as lax
import jax.numpy as jnp

variable = functools.partial(sow, tag="variable")
harvest_variables = functools.partial(harvest, tag="variable")


def body(carry, x):
    def _t(x):
        return variable(x + carry, name="x", mode="clobber")

    def _f(x):
        return variable(x, name="x", mode="clobber")

    x = lax.cond(True, _t, _f, x)
    return x, x


def f(init):
    return lax.scan(jax.jit(body), init, jnp.arange(5.0))


def h(x):
    x = variable(x + 1, name="x", mode="append")
    # x = variable(x + 1, name="x", mode="append")
    return x


def i(x):
    x = nest(h, scope="h1")(x)
    x = nest(h, scope="h2")(x)
    return x


x = reap(i, tag="variable")(1.0)
print(x)



import flarejax as fj
import jax
from jaxtyping import Array, Float

import jax.nn as jnn

__all__ = [
    "ELU",
    "GLU",
    "CeLU",
    "HardSigmoid",
    "HardSiLU",
    "HardTanh",
    "LeakyReLU",
    "LogSigmoid",
    "LogSoftmax",
    "LogSumExp",
    "OneHot",
    "ReLU",
    "ReLU6",
    "SeLU",
    "Sigmoid",
    "SiLU",
    "Softmax",
    "SoftPlus",
    "SoftSign",
    "SparsePlus",
    "SquarePlus",
    "Standardize",
]


class CeLU(fj.Module):
    __module_name = "flarenet.CeLU"

    @fj.typecheck
    def __call__(self, x: Float[Array, "*s"]) -> Float[Array, "*s"]:
        return jnn.relu(x)


class ELU(fj.Module):
    __module_name = "flarenet.ELU"

    @fj.typecheck
    def __call__(self, x: Float[Array, "*s"]) -> Float[Array, "*s"]:
        return jnn.elu(x)


class GELU(fj.Module):
    __module_name = "flarenet.GELU"

    @fj.typecheck
    def __call__(self, x: Float[Array, "*s"]) -> Float[Array, "*s"]:
        return jnn.gelu(x)


class GLU(fj.Module):
    __module_name = "flarenet.GLU"

    axis: int = -1

    @fj.typecheck
    def __call__(self, x: Float[Array, "..."]) -> Float[Array, "..."]:
        return jnn.glu(x, self.axis)


class HardSigmoid(fj.Module):
    __module_name = "flarenet.HardSigmoid"

    @fj.typecheck
    def __call__(self, x: Float[Array, "*s"]) -> Float[Array, "*s"]:
        return jnn.hard_sigmoid(x)


class HardSiLU(fj.Module):
    __module_name = "flarenet.HardSiLU"

    @fj.typecheck
    def __call__(self, x: Float[Array, "*s"]) -> Float[Array, "*s"]:
        return jnn.hard_silu(x)


class HardTanh(fj.Module):
    __module_name = "flarenet.HardTanh"

    @fj.typecheck
    def __call__(self, x: Float[Array, "*s"]) -> Float[Array, "*s"]:
        return jnn.hard_tanh(x)


class LeakyReLU(fj.Module):
    __module_name = "flarenet.LeakyReLU"

    negative_slope: float = 1e-2

    @fj.typecheck
    def __call__(self, x: Float[Array, "*s"]) -> Float[Array, "*s"]:
        return jnn.leaky_relu(x, self.negative_slope)


class LogSigmoid(fj.Module):
    __module_name = "flarenet.LogSigmoid"

    @fj.typecheck
    def __call__(self, x: Float[Array, "*s"]) -> Float[Array, "*s"]:
        return jnn.log_sigmoid(x)


class LogSoftmax(fj.Module):
    __module_name = "flarenet.LogSoftmax"

    @fj.typecheck
    def __call__(self, x: Float[Array, "*s"]) -> Float[Array, "*s"]:
        return jnn.log_softmax(x, axis=-1)


class LogSumExp(fj.Module):
    __module_name = "flarenet.LogSumExp"

    @fj.typecheck
    def __call__(self, x: Float[Array, "..."]) -> Float[Array, "..."]:
        return jnn.logsumexp(x, axis=-1)


class Standardize(fj.Module):
    __module_name = "flarenet.Standardize"

    axis: int = -1

    @fj.typecheck
    def __call__(self, x: Float[Array, "*s"]) -> Float[Array, "*s"]:
        return jnn.standardize(x, axis=self.axis)


class OneHot(fj.Module):
    __module_name = "flarenet.OneHot"

    num_classes: int
    axis: int = -1

    @fj.typecheck
    def __call__(
        self,
        x: Float[Array, "*b"],
    ) -> Float[Array, "*b {self.dim}"]:
        return jnn.one_hot(x, self.num_classes, axis=self.axis)


class ReLU(fj.Module):
    __module_name = "flarenet.ReLU"

    @fj.typecheck
    def __call__(self, x: Float[Array, "*s"]) -> Float[Array, "*s"]:
        return jnn.relu(x)


class ReLU6(fj.Module):
    __module_name = "flarenet.ReLU6"

    @fj.typecheck
    def __call__(self, x: Float[Array, "*s"]) -> Float[Array, "*s"]:
        return jnn.relu6(x)


class SeLU(fj.Module):
    __module_name = "flarenet.SeLU"

    @fj.typecheck
    def __call__(self, x: Float[Array, "*s"]) -> Float[Array, "*s"]:
        return jnn.selu(x)


class Sigmoid(fj.Module):
    __module_name = "flarenet.Sigmoid"

    @fj.typecheck
    def __call__(self, x: Float[Array, "*s"]) -> Float[Array, "*s"]:
        return jnn.sigmoid(x)


class SoftSign(fj.Module):
    __module_name = "flarenet.SoftSign"

    @fj.typecheck
    def __call__(self, x: Float[Array, "*s"]) -> Float[Array, "*s"]:
        return jnn.soft_sign(x)


class Softmax(fj.Module):
    __module_name = "flarenet.Softmax"

    axis: int = -1

    @fj.typecheck
    def __call__(self, x: Float[Array, "*s"]) -> Float[Array, "*s"]:
        return jnn.softmax(x, axis=self.axis)


class SoftPlus(fj.Module):
    __module_name = "flarenet.SoftPlus"

    @fj.typecheck
    def __call__(self, x: Float[Array, "*s"]) -> Float[Array, "*s"]:
        return jnn.softplus(x)


class SparsePlus(fj.Module):
    __module_name = "flarenet.SparsePlus"

    @fj.typecheck
    def __call__(self, x: Float[Array, "*s"]) -> Float[Array, "*s"]:
        return jnn.sparse_plus(x)


class SiLU(fj.Module):
    __module_name = "flarenet.SiLU"

    @fj.typecheck
    def __call__(self, x: Float[Array, "*s"]) -> Float[Array, "*s"]:
        return jnn.silu(x)


class SquarePlus(fj.Module):
    __module_name = "flarenet.SquarePlus"

    @fj.typecheck
    def __call__(self, x: Float[Array, "*s"]) -> Float[Array, "*s"]:
        return jnn.squareplus(x)



# %%
import flarejax as fj
import flarenet as fn

from typing import Any, TypeVar, ParamSpec, Callable, Generic, overload
import inspect

import jax
import jax.numpy as jnp
import jax.random as jrandom


key = jrandom.PRNGKey(0)
key1, key2, key3 = jrandom.split(key, 3)

model = fj.Sequential(
    (
        fn.Linear.init(key, 2, 32),
        fn.GELU(),
        fn.Linear.init(key2, 32, 32),
        fn.GELU(),
        fn.Linear.init(key3, 32, 2),
    )
)
model

# %%
type(model.__call__.__self__)

# %%
T = TypeVar("T")
P = ParamSpec("P")


class Return(fj.Module, Generic[T], register=False):
    self: Any
    hook: Any
    x: T

    def __iter__(self):
        return iter((self.self, self.x, self.hook))


@overload
def apply(f: Callable[P, T], *args: P.args, **kwargs: P.kwargs) -> Return[T]: ...


@overload
def apply(
    f: Callable[P, Return[T]], *args: P.args, **kwargs: P.kwargs
) -> Return[T]: ...


def apply(
    f: Callable[P, T] | Callable[P, Return[T]],
    *args: P.args,
    **kwargs: P.kwargs,
) -> Return[T]:
    signature = inspect.signature(inspect.unwrap(f))
    print(signature)

    r = f(*args, **kwargs)

    if isinstance(r, Return):
        return r

    return Return(
        self=None,
        hook=None,
        x=r,
    )


x = jnp.ones((1, 2))
self, y, hook = apply(model, x)
print(y)

# %%
(lambda x: x)(1, 2)



from typing import TypeVar

import oryx.core.interpreters.harvest as harvest

T = TypeVar("T")


def tag_mode_sow(x: T, /, *, name: str) -> T:
    x = harvest.sow(x, name=name, tag="strict", mode="strict")
    x = harvest.sow(x, name=name, tag="clobber", mode="clobber")
    x = harvest.sow(x, name=name, tag="append", mode="append")
    return x



# %%
import ast
import astor


class AssignmentPrinter(ast.NodeTransformer):
    def visit_FunctionDef(self, node):
        new_body = []
        for stmt in node.body:
            new_body.append(stmt)
            if isinstance(stmt, ast.Assign):
                for target in stmt.targets:
                    if isinstance(target, ast.Name):
                        print_stmt = ast.Expr(
                            value=ast.Call(
                                func=ast.Name(id="print", ctx=ast.Load()),
                                args=[ast.Constant(f"{target.id} ="), target],
                                keywords=[],
                            )
                        )
                        new_body.append(print_stmt)
        node.body = new_body
        return node


def modify_and_print_assignments(source_code):
    # Parse the source code into an AST
    tree = ast.parse(source_code)

    # Transform the AST to add print statements
    transformer = AssignmentPrinter()
    transformed_tree = transformer.visit(tree)

    # Convert the modified AST back to source code
    modified_code = astor.to_source(transformed_tree)

    return modified_code


source_code = """
def calculate_total(price, quantity, tax_rate=0.05, discount=0.1):
    subtotal = price * quantity
    tax = subtotal * tax_rate
    discount_amount = subtotal * discount
    total = subtotal + tax - discount_amount
    return total
"""

modified_code = modify_and_print_assignments(source_code)
print(modified_code)

# Optional: Execute the modified code
exec(modified_code)

# Call the function to see the prints
calculate_total(100, 5)



import ast
import inspect
from ast import fix_missing_locations


def print_nodes(target: ast.Name | ast.Tuple, node: ast.Assign):
    if isinstance(target, ast.Tuple):
        exprs = []

        for t in target.elts:
            if not isinstance(t, ast.Name):
                print(f"Skipping target of type {type(t)}")
                continue

            exprs.extend(print_nodes(t, node))

        return exprs

    if not isinstance(target, ast.Name):
        error = f"Expected target to be an ast.Name, got {type(target)}"
        raise ValueError(error)

    value = ast.Constant(f"variable {target.id} is assigned the value")
    acall = ast.Call(
        func=ast.Name(id="print", ctx=ast.Load()),
        args=[value, target],
        keywords=[],
    )
    expr = ast.Expr(value=acall)
    return [expr]


class AddPrints(ast.NodeTransformer):
    def visit_Assign(self, node):
        new_nodes = []

        for target in node.targets:
            # if isinstance(target, ast.Tuple):
            #     for t in target.elts:
            #         if not isinstance(t, ast.Name):
            #             print(f"Skipping target of type {type(t)}")
            #             continue

            #         new_nodes.append(print_nodes(t, node))

            # if not isinstance(target, ast.Name):
            #     print(f"Skipping target of type {type(target)}")
            #     continue

            if not isinstance(target, ast.Name | ast.Tuple):
                continue

            new_nodes.extend(print_nodes(target, node))

        return [node, *new_nodes]


def add_prints(func):
    # Inspect gets the source, then parse into an AST
    source = inspect.getsource(func)
    tree = ast.parse(source)

    # Transform the AST
    AddPrints().visit(tree)

    # fix location info
    fix_missing_locations(tree)

    # turn transformed AST back into code
    new_func_code = compile(tree, filename="<ast>", mode="exec")

    # Create a new dictionary for the function's local namespace
    new_func_namespace = {}
    exec(new_func_code, new_func_namespace)

    # Return the function from the new namespace
    return new_func_namespace[func.__name__]


# Test function
def test():
    a = 1
    b = 2
    c, d = 3, 4

    *e, f = [1, 2, 3, 4, 5]


# Add two more assignments into the function
test_transformed = add_prints(test)
test_transformed()



import flarejax as fj
import jax
from jaxtyping import Array, Float, jaxtyped

from ._activation import Standardize
from ._linear import Bias, Scale


class LayerNorm(fj.Module):
    __module_name = "flarenet.LayerNorm"

    norm: Standardize
    b: Bias
    s: Scale

    @fj.typecheck
    @classmethod
    def init(cls, dim: int, axis: int = -1):
        return cls(
            norm=Standardize(axis=axis),
            b=Bias.init(dim),
            s=Scale.init(dim),
        )

    @jaxtyped(typechecker=fj.typecheck)
    def __call__(
        self,
        x: Float[Array, "*b {self.dim}"],
    ) -> Float[Array, "*b {self.dim}"]:
        return self.s(self.b(self.norm(x)))

    @fj.typecheck
    @property
    def dim(self) -> int:
        return self.b.dim



def tuple_to_nested_dict(input_dict):
    def insert_nested_dict(d, keys, value):
        if len(keys) == 1:
            d[keys[0]] = value
        else:
            if keys[0] not in d:
                d[keys[0]] = {}
            insert_nested_dict(d[keys[0]], keys[1:], value)

    nested_dict = {}
    for k, v in input_dict.items():
        insert_nested_dict(nested_dict, k, v)

    return nested_dict


# Example usage
input_dict = {("a", "b", "c"): 1, ("a", "b", "d"): 2, ("a", "e"): 3, ("f",): 4}

nested_dict = tuple_to_nested_dict(input_dict)
print(nested_dict)



from typing import Callable, Generator, TypeVar, Any, ParamSpec, Concatenate
import functools

T = TypeVar("T")
P = ParamSpec("P")


def generator_to_function(
    generator: Callable[P, Generator[Any, None, T]], /
) -> Callable[P, T]:
    @functools.wraps(generator)
    def wrapped(*args: P.args, **kwargs: P.kwargs) -> T:
        gen = generator(*args, **kwargs)

        try:
            while True:
                _ = next(gen)

        except StopIteration as e:
            return e.value

    return wrapped


def generator_to_function_maybe(
    generator: Callable[P, Generator[Any, None, T]]
) -> Callable[Concatenate[bool, P], Any | Generator[Any, None, T]]:
    print(type(P.kwargs))
    print(dir(P.kwargs))

    def wrapped(
        *args: P.args,
        **kwargs: P.kwargs | {"as_generator": bool},
    ) -> T | Generator[Any, None, T]:
        if kwargs["as_generator"]:
            return generator(*args, **kwargs)

        gen = generator(*args, **kwargs)

        try:
            while True:
                _ = next(gen)

        except StopIteration as e:
            return e.value

    return wrapped


@generator_to_function_maybe
def g():

    yield 1
    yield 2
    return 3


for i in g(as_generator=True):
    print(i)



import inspect


def get_var_name(value):
    callers_local_vars = inspect.currentframe().f_back.f_locals.items()
    var_name = [
        var_name for var_name, var_val in callers_local_vars if var_val is value
    ]
    return var_name


def example_function(x):
    var_name = get_var_name(x)
    print(f"The variable name is: {var_name[0] if var_name else 'Unknown'}")


# Example usage
a = 10
example_function(a)



import ast
import inspect
from ast import fix_missing_locations

import warnings


def repr_node(node):
    if isinstance(node, ast.Name):
        return node.id

    if isinstance(node, ast.Constant):
        return node.value

    if isinstance(node, ast.Subscript):
        return f"{repr_node(node.value)}[{repr_node(node.slice)}]"

    if isinstance(node, ast.Starred):
        return f"*{repr_node(node.value)}"

    warnings.warn(f"Unknown node type {type(node)}")
    return "<Unknown>"


def print_nodes(target: ast.expr, node):
    value = ast.Constant(f"variable {repr_node(target)}")
    acall = ast.Call(
        func=ast.Name(id="print", ctx=ast.Load()),
        args=[value, node.value],
        keywords=[],
    )
    expr = ast.Expr(value=acall)
    return [expr]


class AddPrints(ast.NodeTransformer):
    def visit_Assign(self, node):
        new_nodes = []
        stack = list(node.targets)

        while stack:
            target = stack.pop()

            if isinstance(target, ast.Tuple):
                # print(type(node))
                # print(dir(target), target.dims)
                stack.extend(target.elts[::-1])
                continue

            new_nodes.extend(print_nodes(target, node))

        return [node, *new_nodes]


def add_prints(func):
    # Inspect gets the source, then parse into an AST
    source = inspect.getsource(func)
    tree = ast.parse(source)

    # Transform the AST
    AddPrints().visit(tree)

    # fix location info
    # print(tree)
    # print(fix_missing_locations(tree), "hello")
    tree = fix_missing_locations(tree)

    # turn transformed AST back into code
    new_func_code = compile(tree, filename="<ast>", mode="exec")

    # Create a new dictionary for the function's local namespace
    new_func_namespace = {}
    exec(new_func_code, new_func_namespace)

    # Return the function from the new namespace
    return new_func_namespace[func.__name__]


# Test function
# @add_prints
def test():
    a = 1
    b = 2
    c, d = 3, 4

    # assert False
    *e, f = [1, 2, 3, 4, 5]
    a = 3

    e[1] = 5


# Add two more assignments into the function
test_transformed = add_prints(test)
test_transformed()



import dataclasses
from typing import Any, Generic, Hashable, Self, TypeVar

import flarejax as fj
import jax
from jaxtyping import PRNGKeyArray


# TODO forget state, if people want statefull computations f them


class Return(fj.Module):
    self: Any
    x: Any
    hook: Any

    def __iter__(self):
        return iter((self.self, self.x, self.hook))


K = TypeVar("K", bound=Hashable)
T = TypeVar("T")


def _nested_dict(input_dict):
    def insert_nested_dict(d, keys, value):
        if len(keys) == 1:
            d[keys[0]] = value
        else:
            if keys[0] not in d:
                d[keys[0]] = {}
            insert_nested_dict(d[keys[0]], keys[1:], value)

    nested_dict = {}
    for k, v in input_dict.items():
        insert_nested_dict(nested_dict, k, v)

    return nested_dict


class Hook(fj.Module, Generic[K, T]):
    _key: PRNGKeyArray
    _context: tuple[K, ...] = fj.field(default=(), static=True)
    _watched: fj.ModuleMapping[tuple[K, ...], T] = fj.field(
        default_factory=fj.ModuleMapping
    )

    @property
    def context(self) -> tuple[K, ...]:
        return self._context

    @property
    def watched(self) -> dict:
        return _nested_dict(self._watched._data)

    def get_key(self) -> tuple[Self, PRNGKeyArray]:
        key, subkey = jax.random.split(self._key)
        return dataclasses.replace(self, _key=subkey), key

    def add(self: Self, x: "Hook[K, T]", /) -> Self:
        watched = self._watched.update(x._watched._data)
        return dataclasses.replace(self, _watched=watched)

    def sub(self: Self, c: K, /):
        return dataclasses.replace(self, _context=(*self._context, c))

    def __call__(self: Self, k: K, v: Any, /) -> tuple[Self, Any]:
        watch = self.watch(k, v)

        if watch is not None:
            watched = self._watched.update({(*self._context, k): v})
            self = dataclasses.replace(self, _watched=watched)

        return self.plant(k, v), self

    def watch(self, k: K, v: Any, /) -> Any:
        return v

    def plant(self, k: K, v: Any, /) -> Self:
        return v


class Linear(fj.Module):
    # The __init__ method is automatically generated
    w: jax.Array
    b: jax.Array

    # additional intialization methods via classmethods
    @classmethod
    def init(cls, key, dim_in, dim):
        w = jax.random.normal(key, (dim_in, dim)) * 0.02
        b = jax.numpy.zeros((dim,))
        return cls(w=w, b=b)

    def __call__(self, x, *, hook: Hook):
        y = x @ self.w
        y, hook = hook("x @ w", y)

        y = y + self.b
        y, hook = hook("y + b", y)

        return Return(self, y, hook)


class Sequential(fj.ModuleSequence):
    __module_name = "flarenet.Sequential"

    def __call__(self, x, *, hook: Hook):
        for i, layer in enumerate(self):
            x = layer(x, hook=hook.sub(f"{self.__class__.__name__}:{i}"))

            if isinstance(x, Return):
                self, x, hook_new = x
                hook = hook.add(hook_new)

        return Return(self, x, hook)


# model = Linear.init(jax.random.PRNGKey(42), 3, 2)
model = Sequential(
    (
        Linear.init(jax.random.PRNGKey(42), 3, 2),
        Linear.init(jax.random.PRNGKey(42), 2, 5),
    )
)

hook = Hook(jax.random.PRNGKey(42))

x = jax.numpy.array([1.0, 2.0, 3.0])
self, y, hook = model(x, hook=hook)


from pprint import pprint

pprint(hook.watched)



class Point:
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __iter__(self):
        yield self.x
        yield self.y


# Example usage
point = Point(1, 2)
x, y = point
print(x)  # Output: 1
print(y)  # Output: 2



import math

import flarejax as fj
import jax.numpy as jnp
import jax.random as jrandom
from jaxtyping import Array, Float, PRNGKeyArray, jaxtyped

from ._utils import tag_mode_sow

__all__ = [
    "Linear",
    "Bias",
    "Scale",
    "Constant",
]


class Linear(fj.Module):
    """
    Apply a learanable affine transformation to the input data. That takes the
    form of a matrix multiplication and an optional bias addition.

    The last input axis has dimensionality 'dim_in', the output shares all axes,
    except the last one with the input, which has dimensionality 'dim'.

    Attributes:
    ---
    w: jax.Array
        The learnable weights of the layer. This has shape (dim_in, dim).

    b: jax.Array | None
        The learnable bias of the layer. This has shape (dim,) or is None, if
        the layer does not have a bias.
    """

    __module_name = "flarenet.Linear"

    w: Float[Array, "dim_in dim"]
    b: Float[Array, "dim"] | None

    @fj.typecheck
    @classmethod
    def init(
        cls,
        key: PRNGKeyArray,
        dim_in: int,
        dim: int,
        use_bias: bool = True,
    ):
        """
        Initialize the layer with random weights and biases. The default
        initialization is based on the Glorot uniform initialization, which
        is the same as the PyTorch default.

        Parameters:
        ---
        key: PRNGKey
            The random key to use for initialization.

        dim_in: int
            The number of input features.

        dim: int
            The number of output features.

        use_bias: bool
            Whether to use a bias term in the layer.

        Returns:
        ---
        Linear
            The initialized layer.
        """
        scale = 1 / math.sqrt(dim_in)
        w_key, b_key = jrandom.split(key)

        w = jrandom.uniform(w_key, (dim_in, dim), minval=-1, maxval=1) * scale

        if use_bias:
            b = jrandom.uniform(b_key, (dim,), minval=-1, maxval=1) * scale
        else:
            b = None

        return cls(w=w, b=b)

    @jaxtyped(typechecker=fj.typecheck)
    def __call__(
        self,
        x: Float[Array, "*b {self.dim_in}"],
    ) -> Float[Array, "*b {self.dim}"]:
        y = jnp.dot(x, self.w)

        y = tag_mode_sow(y, name="x @ w")

        if self.has_bias:
            y += self.b
            y = tag_mode_sow(y, name="x @ w + b")

        return y

    @fj.typecheck
    @property
    def dim_in(self) -> int:
        return self.w.shape[0]

    @fj.typecheck
    @property
    def dim(self) -> int:
        return self.w.shape[1]

    @fj.typecheck
    @property
    def has_bias(self) -> bool:
        return self.b is not None


class Bias(fj.Module):
    __module_name = "flarenet.Bias"

    b: Float[Array, "..."]

    @fj.typecheck
    @classmethod
    def init(cls, dim: int):
        return cls(b=jnp.zeros((dim,)))

    @jaxtyped(typechecker=fj.typecheck)
    def __call__(
        self,
        x: Float[Array, "*b {self.dim}"],
    ) -> Float[Array, "*b {self.dim}"]:
        y = x + self.b
        y = tag_mode_sow(y, name="x + b")
        return y

    @fj.typecheck
    @property
    def dim(self) -> int:
        return self.b.shape[0]


class Scale(fj.Module):
    __module_name = "flarenet.Scale"

    s: Float[Array, "..."]

    @classmethod
    def init(cls, dim: int):
        return cls(s=jnp.ones((dim,)))

    @jaxtyped(typechecker=fj.typecheck)
    def __call__(
        self, x: Float[Array, "*b {self.dim}"]
    ) -> Float[Array, "*b {self.dim}"]:
        y = x * self.s
        y = tag_mode_sow(y, name="x * s")
        return y

    @property
    def dim(self) -> int:
        return self.s.shape[0]


class Constant(fj.Module):
    __module_name = "flarenet.Constant"

    x: Float[Array, "..."]

    @classmethod
    def random_normal(
        cls,
        key: PRNGKeyArray,
        shape: tuple[int, ...],
        std: float = 1.0,
    ):
        return cls(x=jrandom.normal(key, shape) * std)

    @classmethod
    def random_uniform(
        cls,
        key: PRNGKeyArray,
        shape: tuple[int, ...],
        low: float = -1.0,
        high: float = 1.0,
    ):
        return cls(x=jrandom.uniform(key, shape, minval=low, maxval=high))

    @classmethod
    def full(cls, x: float = 0.0, shape: tuple[int, ...] = ()):
        return cls(x=jnp.full(shape, x))

    @property
    def shape(self) -> tuple[int, ...]:
        return self.x.shape

    @property
    def dtype(self) -> jnp.dtype:
        return self.x.dtype

    @fj.typecheck
    def __call__(self, *args, **kwargs) -> Array:
        return self.x



from typing import Any, TypeVar, ParamSpec, Callable, Generic, overload

import flarejax as fj
import inspect


T = TypeVar("T")
P = ParamSpec("P")


class Return(fj.Module, Generic[T]):
    self: Any
    hook: Any
    x: T

    def __iter__(self):
        return iter((self.self, self.x, self.hook))


@overload
def apply(f: Callable[P, T], *args: P.args, **kwargs: P.kwargs) -> Return[T]: ...


@overload
def apply(
    f: Callable[P, Return[T]], *args: P.args, **kwargs: P.kwargs
) -> Return[T]: ...


def apply(
    f: Callable[P, T] | Callable[P, Return[T]],
    *args: P.args,
    **kwargs: P.kwargs,
) -> Return[T]:
    r = f(*args, **kwargs)

    if isinstance(r, Return):
        return r

    return Return(
        self=None,
        hook=None,
        x=r,
    )



import dis
import functools
import jax

import sys


def trace_assignments(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        bytecode = dis.Bytecode(func)
        assignment_ops = {"STORE_NAME", "STORE_FAST", "STORE_GLOBAL"}

        def trace(frame, event, _):
            if event == "line":
                for instr in bytecode:
                    if instr.opname in assignment_ops:
                        varname = instr.argval
                        if varname in frame.f_locals:
                            print(f"Assignment: {varname} = {frame.f_locals[varname]}")

            return trace

        sys.settrace(trace)
        try:
            result = func(*args, **kwargs)
        finally:
            sys.settrace(None)
        return result

    return wrapper


# Example usage
@trace_assignments
@jax.vmap
def example_function(x):
    # x = 10
    y = 5
    z = x + y
    return z


example_function(jax.numpy.zeros((32, 3)))



import ast
import inspect


# Step 1: Define the original function
def original_function(x, y):
    assert False
    return x + y


# Step 2: Parse the function's source code into an AST
source_code = inspect.getsource(original_function)
tree = ast.parse(source_code)


# Step 3: Modify the AST
class Transformer(ast.NodeTransformer):
    def visit_BinOp(self, node):
        if isinstance(node.op, ast.Add):
            node.op = ast.Sub()
        return node


transformer = Transformer()
modified_tree = transformer.visit(tree)
ast.fix_missing_locations(modified_tree)

# Step 4: Compile and execute the modified AST
code = compile(modified_tree, filename="<ast>", mode="exec")
exec(code)

# Step 5: Test the modified function
result = original_function(10, 5)
print(result)  # Output should be 5 instead of 15

# Crozet



from typing import Any

import flarejax as fj
import jax


class Hook(fj.Module):
    context: tuple
    key: jax.Array
    mode: Any

    def merge(self, ctx, other):
        pass


class Linear(fj.Module):
    # The __init__ method is automatically generated
    w: jax.Array
    b: jax.Array

    # additional intialization methods via classmethods
    @classmethod
    def init(cls, key, dim_in, dim):
        w = jax.random.normal(key, (dim, dim_in)) * 0.02
        b = jax.numpy.zeros((dim,))
        return cls(w=w, b=b)

    def __call__(self, x, *, hook=None):
        return self.w @ x + self.b


key = jax.random.PRNGKey(42)
key1, key2 = jax.random.split(key)

model = fj.Sequential(
    (
        Linear.init(key1, 3, 2),
        Linear.init(key2, 2, 5),
    )
)



from typing import Generator, Any, TypeVar, Callable, ParamSpec
import functools
import inspect

T = TypeVar("T")
P = ParamSpec("P")


def generator_to_function(
    generator: Callable[P, Generator[Any, T, T]] | Callable[P, T],
    /,
) -> Callable[P, T]:
    @functools.wraps(generator)
    def wrapped(*args: P.args, **kwargs: P.kwargs) -> T:
        gen = generator(*args, **kwargs)

        if not inspect.isgenerator(gen):
            return gen  # type: ignore

        try:
            x, _ = next(gen)

            while True:
                x, _ = gen.send(x)

        except StopIteration as e:
            return e.value

    return wrapped


def function_to_generator(
    function: Callable[P, T],
    /,
) -> Callable[P, Generator[tuple[T, tuple[str]], T, T]]:
    @functools.wraps(function)
    def wrapped(
        *args: P.args, **kwargs: P.kwargs
    ) -> Generator[tuple[T, tuple[str]], T, T]:

        x = function(*args, **kwargs)
        x = yield x, ("FunctionToGenerator",)
        return x

    return wrapped


class Bias:
    def __call__(self, x):
        x = yield x + 1, ("Bias",)
        return x


class Linear:
    def __call__(self, x):
        x = yield x * 2, ("Linear",)
        return x


class Affine:
    def __init__(self):
        self.bias = Bias()
        self.linear = Linear()

    def __call__(self, x):
        x = add_info(self.linear(x), "Affine")
        x = yield from x
        x = yield from add_info(self.bias(x), "Affine")
        return x


class Sequential:
    def __init__(self, *layers):
        self.layers = layers

    def __call__(self, x):
        for layer in self.layers:
            # print(x)
            x = yield from add_info(layer(x), "Sequential")

        return x


def add_info(gen, info):
    try:
        x, info_old = next(gen)
        x = yield (x, (info,) + info_old)

        while True:
            x, info_old = gen.send(x)
            x = yield (x, (info,) + info_old)

    except StopIteration as e:
        return e.value


model = Sequential(Affine(), Affine())
model = generator_to_function(model)

print(model(1))



from ._activation import (
    ELU,
    GELU,
    GLU,
    CeLU,
    HardSigmoid,
    HardSiLU,
    HardTanh,
    LeakyReLU,
    LogSigmoid,
    LogSoftmax,
    LogSumExp,
    OneHot,
    ReLU,
    ReLU6,
    SeLU,
    Sigmoid,
    SiLU,
    Softmax,
    SoftPlus,
    SoftSign,
    SparsePlus,
    SquarePlus,
    Standardize,
)
from ._combine import (
    Add,
    Concat,
    Identity,
    Index,
    Multiply,
    Residual,
    Sequential,
)
from ._linear import Bias, Constant, Linear, Scale
from ._norm import LayerNorm

__version__ = "0.3.11"

__all__ = [
    "ELU",
    "GELU",
    "GLU",
    "CeLU",
    "HardSigmoid",
    "HardSiLU",
    "HardTanh",
    "LeakyReLU",
    "LogSigmoid",
    "LogSoftmax",
    "LogSumExp",
    "OneHot",
    "ReLU",
    "ReLU6",
    "SeLU",
    "Sigmoid",
    "SiLU",
    "Softmax",
    "SoftPlus",
    "SoftSign",
    "SparsePlus",
    "SquarePlus",
    "Standardize",
    "Add",
    "Concat",
    "Identity",
    "Index",
    "Multiply",
    "Residual",
    "Sequential",
    "Bias",
    "Constant",
    "Linear",
    "Scale",
    "LayerNorm",
    "__version__",
]



from greenlet import greenlet


class Bias:
    def __call__(self, x):
        return x + 1


class Linear:
    def __call__(self, x):
        return x * 2


class Affine:
    def __init__(self):
        self.bias = Bias()
        self.linear = Linear()

    def __call__(self, x):
        x = self.linear(x)
        x = self.bias(x)
        return x


class Sequential:
    def __init__(self, *layers):
        self.layers = layers

    def __call__(self, x):
        for layer in self.layers:
            x = greenlet.getcurrent().parent.switch(x)
            print(x)
            x = layer(x)

        print("done")
        return x


model = Sequential(Affine(), Affine())

green = greenlet(model)
r = green.switch("hello")
r = green.switch(1)

print(r)



from typing import Generator


def f():
    yield 1
    x = yield 2
    return x


def g():
    x = (yield from f()) if isinstance(f(), Generator) else f()
    return x


def run_generator(gen):
    try:
        x = next(gen)
        print("received", x)

        while True:
            # print("sending None")
            x = gen.send("message")
            print("received", x)
    except StopIteration as e:
        print("return value", e.value)


r = run_generator(g())



from setuptools import setup, find_packages
from flarenet import __version__


with open("README.md", "r") as fh:
    long_description = fh.read()


with open("requirements.txt") as f:
    requirements = f.read().splitlines()


description = "Common neural network modules and utilities based on flarejax."

setup(
    name="flarenet",
    version=__version__,
    packages=find_packages(where="."),
    python_requires=">=3.10",
    install_requires=requirements,
    description=description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pwolle/flarenet",
    author="Paul Wollenhaupt",
    author_email="paul.wollenhaupt@gmail.com",
)



# TODO

from typing import Any
import flarejax as fj

from jaxtyping import Array, Float


class Conv2D(fj.Module):
    w: Float[Array, "dim dim_in h w"]
    b: Float[Array, "dim"]

    @classmethod
    def init(cls):
        raise NotImplementedError

    def __call__(self, x):
        raise NotImplementedError



from typing import Callable, Generator, TypeVar, Any, ParamSpec
import inspect
import functools

T = TypeVar("T")
P = ParamSpec("P")


def run_generator(
    g: Callable[P, Generator[Any, None, T]],
    /,
    *args: P.args,
    **kwargs: P.kwargs,
) -> T:
    try:
        gen = g(*args, **kwargs)

        while True:
            _ = next(gen)

    except StopIteration as e:
        return e.value


def generator_to_function(
    generator: Callable[P, Generator[Any, None, T]] | Callable[P, T], /
) -> Callable[P, T]:
    if not inspect.isgeneratorfunction(generator):
        return generator  # type: ignore

    @functools.wraps(generator)
    def wrapped(*args: P.args, **kwargs: P.kwargs) -> T:
        gen = generator(*args, **kwargs)

        try:
            while True:
                _ = next(gen)

        except StopIteration as e:
            return e.value

    return wrapped



def run_generator(gen):
    try:
        x = next(gen)
        print("received", x)

        while True:
            # print("sending None")
            x = gen.send("message")
            print("received", x)
    except StopIteration as e:
        print("return value", e.value)


def f1():
    print("yielding 1 from f1")
    x = yield 1
    print("got", x)
    print("yielding 2 from f2")
    yield 2
    return 3


def f2(x):
    print("yielding 3 from f2")
    yield 3
    print("yielding 4 from f2")
    yield 4


def f3():
    a = yield from f1()
    print(f"{a=}")
    return a


# def generator_to_function(gen):


r = run_generator(f3())
# print(r)



from typing import Any, TypeVar

import flarejax as fj
import jax
import jax.numpy as jnp
import oryx.core.interpreters.harvest as harvest

T = TypeVar("T")


__all__ = [
    "Add",
    "Multiply",
    "Concat",
    "Identity",
    "Residual",
    "Index",
]


class Sequential(fj.ModuleSequence):
    __module_name = "flarenet.Sequential"

    @fj.typecheck
    def __call__(self, x, *args, **kwargs) -> Any:
        for i, layer in enumerate(self):
            assert callable(layer), f"Layer {layer} is not callable."

            # layer = harvest.nest(layer, scope=str(i))
            x = layer(x, *args, **kwargs)

        return x


class Add(Sequential):
    __module_name = "flarenet.Add"

    @fj.typecheck
    def __call__(self, x, *args, **kwargs) -> Any:
        y = 0

        for i, layer in enumerate(self):
            assert callable(layer), f"Layer {layer} is not callable."

            layer = harvest.nest(layer, scope=str(i))
            y = y + layer(x, *args, **kwargs)

        return y


class Multiply(Sequential):
    __module_name = "flarenet.Multiply"

    @fj.typecheck
    def __call__(self, x, *args, **kwargs) -> Any:
        y = 1

        for i, layer in enumerate(self):
            assert callable(layer), f"Layer {layer} is not callable."

            layer = harvest.nest(layer, scope=str(i))
            y = y * layer(x, *args, **kwargs)

        return y


class Concat(Sequential):
    __module_name = "flarenet.Concat"

    @fj.typecheck
    def __call__(self, x, *args, **kwargs) -> jax.Array:
        ys = []

        for i, layer in enumerate(self):
            assert callable(layer), f"Layer {layer} is not callable."

            layer = harvest.nest(layer, scope=str(i))
            ys.append(layer(x, *args, **kwargs))

        return jnp.concatenate(ys, axis=-1)


class Identity(fj.Module):
    __module_name = "flarenet.Identity"

    @fj.typecheck
    def __call__(self, x: T, *args, **kwargs) -> T:
        return x


class Residual(fj.Module):
    __module_name = "flarenet.Residual"

    layer: fj.Module

    @fj.typecheck
    def __call__(self, x, *args, **kwargs) -> Any:
        assert callable(self.layer), f"Layer {self.layer} is not callable."
        return x + self.layer(x, *args, **kwargs)


class Index(fj.Module):
    __module_name = "flarenet.Index"

    index: int | slice | tuple[int | slice, ...]

    @fj.typecheck
    def __call__(self, x) -> Any:
        return x[self.index]
